/*
   Class Name:

      CBspLevel.

   Created by:

      Allen Sherrod (Programming Ace of www.UltimateGameProgramming.com).

   Description:

      This class represents a Quake3 3D map that can be rendered in OpenGL
      or in Direct3D.
*/


#include<stdio.h>
#include<string.h>
#include"CBspLevel.h"


bool CBspLevel::LoadLevel(char *filename)
{
   stQ3Header Header;
   stQ3FileLump Lumps[iMaxLumpsOffset];

   // Open file.
   FILE *fp = fopen(filename, "rb");
   if(!fp) return false;

   // Read the header data and make sure this is a bsp level.
   fread(&Header, 1, sizeof(stQ3Header), fp);
   if(strcmp(Header.bspID, "IBSP.") != 0) return false;

   // Read all the lumps in the file.
	fread(&Lumps, iMaxLumpsOffset, sizeof(stQ3FileLump), fp);

   // Allocate data for the vertex points, triangles, textures, lightmaps, etc.
   m_totalVertices = Lumps[iVerticesOffset].length / sizeof(stQ3Vertex);
	m_vertices = new stQ3Vertex[m_totalVertices];

   m_totalFaces = Lumps[iFacesOffset].length / sizeof(stQ3Face);
	m_faces = new stQ3Face[m_totalFaces];

   m_totalTextures = Lumps[iTextureOffset].length / sizeof(stQ3Texture);
   m_textures = new stQ3Texture[m_totalTextures];

   m_totalLightMaps = Lumps[iLightMapsOffset].length / sizeof(stQ3LightMap);
   m_lightMaps = new stQ3LightMap[m_totalLightMaps];

   m_totalNodes = Lumps[iNodesOffset].length / sizeof(stQ3BspNode);
   m_rootNode = new stQ3BspNode[m_totalNodes];

   m_totalLeaves = Lumps[iLeafOffset].length / sizeof(stQ3BspLeaf);
   m_leaves = new stQ3BspLeaf[m_totalLeaves];

   m_totalLeafFaces = Lumps[iLeafFacesOffset].length / sizeof(int);
   m_leafFaces = new int[m_totalLeafFaces];
   
   m_totalPlanes = Lumps[iPlanesOffset].length / sizeof(CPlane);
   m_planes = new CPlane[m_totalPlanes];

   // Load the vertex data.
	fseek(fp, Lumps[iVerticesOffset].offset, SEEK_SET);
	fread(m_vertices, m_totalVertices, sizeof(stQ3Vertex), fp);

   // Load the triangle data.
   fseek(fp, Lumps[iFacesOffset].offset, SEEK_SET);
	fread(m_faces, m_totalFaces, sizeof(stQ3Face), fp);

   // Load the texture data.
	fseek(fp, Lumps[iTextureOffset].offset, SEEK_SET);
	fread(m_textures, m_totalTextures, sizeof(stQ3Texture), fp);

   // Load the lightmap data.
	fseek(fp, Lumps[iLightMapsOffset].offset, SEEK_SET);
	fread(m_lightMaps, m_totalLightMaps, sizeof(stQ3LightMap), fp);

	// Load the node data.
	fseek(fp, Lumps[iNodesOffset].offset, SEEK_SET);
	fread(m_rootNode, m_totalNodes, sizeof(stQ3BspNode), fp);

   // Load the leaf data.
	fseek(fp, Lumps[iLeafOffset].offset, SEEK_SET);
	fread(m_leaves, m_totalLeaves, sizeof(stQ3BspLeaf), fp);

   // Load the leaf faces data.
	fseek(fp, Lumps[iLeafFacesOffset].offset, SEEK_SET);
	fread(m_leafFaces, m_totalLeafFaces, sizeof(int), fp);

   // Load the plane data.
	fseek(fp, Lumps[iPlanesOffset].offset, SEEK_SET);
	fread(m_planes, m_totalPlanes, sizeof(CPlane), fp);

   // If any visibility data, load it.  
   if(Lumps[iVisibleDataOffset].length)
      {
         fseek(fp, Lumps[iVisibleDataOffset].offset, SEEK_SET);
         fread(&m_clusters.totalClusters, 1, sizeof(int), fp);
         fread(&m_clusters.size, 1, sizeof(int), fp);

         m_clusters.bitSet = new unsigned char[m_clusters.totalClusters *
                                               m_clusters.size];
         fread(m_clusters.bitSet, 1, sizeof(unsigned char) *
               (m_clusters.totalClusters * m_clusters.size), fp);
      }

   fclose(fp);
   return true;
}


void CBspLevel::SwapAxis()
{
   float tempf = 0;
   int tempi = 0;

   // Must swap the axis for OpenGL.
   for(int i = 0; i < m_totalVertices; i++)
	   {
		   tempf = m_vertices[i].vertex.y;
		   m_vertices[i].vertex.y = m_vertices[i].vertex.z;
		   m_vertices[i].vertex.z = -tempf;

		   m_vertices[i].texCoord.tv *= -1;
	   }

   for(i = 0; i < m_totalLeaves; i++)
	   {
		   tempi = m_leaves[i].aabbMin.y;
		   m_leaves[i].aabbMin.y = m_leaves[i].aabbMin.z;
		   m_leaves[i].aabbMin.z = -tempi;

         tempi = m_leaves[i].aabbMax.y;
		   m_leaves[i].aabbMax.y = m_leaves[i].aabbMax.z;
		   m_leaves[i].aabbMax.z = -tempi;
	   }

   for(i = 0; i < m_totalPlanes; i++)
	   {
		   tempf = m_planes[i].b;
		   m_planes[i].b = m_planes[i].c;
		   m_planes[i].c = -tempf;
	   }
}


int CBspLevel::GetCameraLeaf(CVector4 &camPos)
{
   int i = 0, result = 0;

   // Loop through all the nodes until we find one that is a negative
   // number.  That negative number means we are in a leaf.
   do
      {
         // Get if the point is in front or behind the plane.
         result = m_planes[m_rootNode[i].planeIndex].ClassifyPoint(camPos);

         // Test if the camera is in front or behind the plane.
         if(result == FRONT_PLANE) i = m_rootNode[i].frontIndex;
         else i = m_rootNode[i].backIndex;
      } while(i >= 0);

   // Since we can't have a negative index, make it positive.
   return -(i + 1);
}


int CBspLevel::isVisible(int cluster1Id, int cluster2Id)
{
   if(!m_clusters.bitSet) return true;
   if(cluster1Id < 0) return true;

   // Get the 8 bit visiblity for this cluster and the one we are testing for.
   unsigned char set = m_clusters.bitSet[(cluster1Id * m_clusters.size) +
                                         (cluster2Id / 8)];
   // Return if this set is visible or not.  We bit shift to get
   // either true or false.
   return set & (1 << (cluster2Id & 7));
}


void CBspLevel::Shutdown()
{
   if(m_lightMaps)
      {
         delete[] m_lightMaps;
         m_lightMaps = NULL;
      }

   if(m_textures)
      {
         delete[] m_textures;
         m_textures = NULL;
      }

   if(m_vertices)
      {
         delete[] m_vertices;
         m_vertices = NULL;
      }

   if(m_faces)
      {
         delete[] m_faces;
         m_faces = NULL;
      }

   if(m_planes)
      {
         delete[] m_planes;
         m_planes = NULL;
      }

   if(m_rootNode)
      {
         delete[] m_rootNode;
         m_rootNode = NULL;
      }

   if(m_leafFaces)
      {
         delete[] m_leafFaces;
         m_leafFaces = NULL;
      }

   if(m_leaves)
      {
         delete[] m_leaves;
         m_leaves = NULL;
      }

   m_totalLightMaps = 0;
   m_totalTextures = 0;
   m_totalVertices = 0;
   m_totalFaces = 0;
   m_totalPlanes = 0;
   m_totalNodes = 0;
   m_totalLeaves = 0;
   m_totalLeafFaces = 0;
}


// Copyright December 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com